# 表格训练


``` python
! [ -e /content ] && pip install -Uqq fastai  # 在Colab上升级fastai
```

> 如何在 fastai 中使用表格应用程序

为了说明表格应用，我们将使用[成人数据集](https://archive.ics.uci.edu/ml/datasets/Adult)的示例，在该示例中，我们需要预测一个人年收入是否超过或低于$50,000，基于一些一般数据。

``` python
from fastai.tabular.all import *
```

我们可以使用常规的 `untar_data` 命令下载该数据集的样本：

``` python
path = untar_data(URLs.ADULT_SAMPLE)
path.ls()
```

    (#3) [Path('/home/ml1/.fastai/data/adult_sample/models'),Path('/home/ml1/.fastai/data/adult_sample/export.pkl'),Path('/home/ml1/.fastai/data/adult_sample/adult.csv')]

然后我们可以看看数据是如何结构化的：

``` python
df = pd.read_csv(path/'adult.csv')
df.head()
```

<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }
&#10;    .dataframe tbody tr th {
        vertical-align: top;
    }
&#10;    .dataframe thead th {
        text-align: right;
    }
</style>

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: right;">
<th data-quarto-table-cell-role="th"></th>
<th data-quarto-table-cell-role="th">age</th>
<th data-quarto-table-cell-role="th">workclass</th>
<th data-quarto-table-cell-role="th">fnlwgt</th>
<th data-quarto-table-cell-role="th">education</th>
<th data-quarto-table-cell-role="th">education-num</th>
<th data-quarto-table-cell-role="th">marital-status</th>
<th data-quarto-table-cell-role="th">occupation</th>
<th data-quarto-table-cell-role="th">relationship</th>
<th data-quarto-table-cell-role="th">race</th>
<th data-quarto-table-cell-role="th">sex</th>
<th data-quarto-table-cell-role="th">capital-gain</th>
<th data-quarto-table-cell-role="th">capital-loss</th>
<th data-quarto-table-cell-role="th">hours-per-week</th>
<th data-quarto-table-cell-role="th">native-country</th>
<th data-quarto-table-cell-role="th">salary</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td data-quarto-table-cell-role="th">0</td>
<td>49</td>
<td>Private</td>
<td>101320</td>
<td>Assoc-acdm</td>
<td>12.0</td>
<td>Married-civ-spouse</td>
<td>NaN</td>
<td>Wife</td>
<td>White</td>
<td>Female</td>
<td>0</td>
<td>1902</td>
<td>40</td>
<td>United-States</td>
<td>&gt;=50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">1</td>
<td>44</td>
<td>Private</td>
<td>236746</td>
<td>Masters</td>
<td>14.0</td>
<td>Divorced</td>
<td>Exec-managerial</td>
<td>Not-in-family</td>
<td>White</td>
<td>Male</td>
<td>10520</td>
<td>0</td>
<td>45</td>
<td>United-States</td>
<td>&gt;=50k</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">2</td>
<td>38</td>
<td>Private</td>
<td>96185</td>
<td>HS-grad</td>
<td>NaN</td>
<td>Divorced</td>
<td>NaN</td>
<td>Unmarried</td>
<td>Black</td>
<td>Female</td>
<td>0</td>
<td>0</td>
<td>32</td>
<td>United-States</td>
<td>&lt;50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">3</td>
<td>38</td>
<td>Self-emp-inc</td>
<td>112847</td>
<td>Prof-school</td>
<td>15.0</td>
<td>Married-civ-spouse</td>
<td>Prof-specialty</td>
<td>Husband</td>
<td>Asian-Pac-Islander</td>
<td>Male</td>
<td>0</td>
<td>0</td>
<td>40</td>
<td>United-States</td>
<td>&gt;=50k</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">4</td>
<td>42</td>
<td>Self-emp-not-inc</td>
<td>82297</td>
<td>7th-8th</td>
<td>NaN</td>
<td>Married-civ-spouse</td>
<td>Other-service</td>
<td>Wife</td>
<td>Black</td>
<td>Female</td>
<td>0</td>
<td>0</td>
<td>50</td>
<td>United-States</td>
<td>&lt;50k</td>
</tr>
</tbody>
</table>

</div>

有些列是连续的（例如年龄），我们将它们视为浮点数，以便直接输入我们的模型。其他列是分类的（例如工作类别或教育），我们会将它们转换为一个唯一的索引，以便输入到嵌入层中。我们可以在`TabularDataLoaders`工厂方法中指定我们的分类和连续列名称，以及因变量的名称：

``` python
dls = TabularDataLoaders.from_csv(path/'adult.csv', path=path, y_names="salary",
    cat_names = ['workclass', 'education', 'marital-status', 'occupation', 'relationship', 'race'],
    cont_names = ['age', 'fnlwgt', 'education-num'],
    procs = [Categorify, FillMissing, Normalize])
```

最后一部分是我们对数据应用的预处理器列表：

- `Categorify`
  将处理每个分类变量，并创建一个从整数到唯一类别的映射，然后用相应的索引替换值。
- `FillMissing`
  将通过现有值的中位数填充连续变量中的缺失值（如果需要，你可以选择一个特定的值）。
- `Normalize` 将对连续变量进行归一化（减去均值并除以标准差）。

为了进一步揭示表面下发生的事情，下面我们将重新编写这个示例，利用
`fastai` 的 `TabularPandas`
类。我们需要进行一个调整，即定义我们想要如何划分数据。默认情况下，上述工厂方法使用了随机的
80/20 划分，因此我们也将采取相同的做法：

``` python
splits = RandomSplitter(valid_pct=0.2)(range_of(df))
```

``` python
to = TabularPandas(df, procs=[Categorify, FillMissing,Normalize],
                   cat_names = ['workclass', 'education', 'marital-status', 'occupation', 'relationship', 'race'],
                   cont_names = ['age', 'fnlwgt', 'education-num'],
                   y_names='salary',
                   splits=splits)
```

一旦我们构建了我们的 `TabularPandas`
对象，我们的数据就完全预处理，如下所示：

``` python
to.xs.iloc[:2]
```

<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }
&#10;    .dataframe tbody tr th {
        vertical-align: top;
    }
&#10;    .dataframe thead th {
        text-align: right;
    }
</style>

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: right;">
<th data-quarto-table-cell-role="th"></th>
<th data-quarto-table-cell-role="th">workclass</th>
<th data-quarto-table-cell-role="th">education</th>
<th data-quarto-table-cell-role="th">marital-status</th>
<th data-quarto-table-cell-role="th">occupation</th>
<th data-quarto-table-cell-role="th">relationship</th>
<th data-quarto-table-cell-role="th">race</th>
<th data-quarto-table-cell-role="th">education-num_na</th>
<th data-quarto-table-cell-role="th">age</th>
<th data-quarto-table-cell-role="th">fnlwgt</th>
<th data-quarto-table-cell-role="th">education-num</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td data-quarto-table-cell-role="th">15780</td>
<td>2</td>
<td>16</td>
<td>1</td>
<td>5</td>
<td>2</td>
<td>5</td>
<td>1</td>
<td>0.984037</td>
<td>2.210372</td>
<td>-0.033692</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">17442</td>
<td>5</td>
<td>12</td>
<td>5</td>
<td>8</td>
<td>2</td>
<td>5</td>
<td>1</td>
<td>-1.509555</td>
<td>-0.319624</td>
<td>-0.425324</td>
</tr>
</tbody>
</table>

</div>

现在我们可以再次构建我们的 `DataLoaders` 了：

``` python
dls = to.dataloaders(bs=64)
```

> 稍后我们将探讨为何使用 `TabularPandas` 进行预处理是有价值的。

`show_batch` 方法的工作方式与其他应用程序类似：

``` python
dls.show_batch()
```

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: right;">
<th data-quarto-table-cell-role="th"></th>
<th data-quarto-table-cell-role="th">workclass</th>
<th data-quarto-table-cell-role="th">education</th>
<th data-quarto-table-cell-role="th">marital-status</th>
<th data-quarto-table-cell-role="th">occupation</th>
<th data-quarto-table-cell-role="th">relationship</th>
<th data-quarto-table-cell-role="th">race</th>
<th data-quarto-table-cell-role="th">education-num_na</th>
<th data-quarto-table-cell-role="th">age</th>
<th data-quarto-table-cell-role="th">fnlwgt</th>
<th data-quarto-table-cell-role="th">education-num</th>
<th data-quarto-table-cell-role="th">salary</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td data-quarto-table-cell-role="th">0</td>
<td>State-gov</td>
<td>Bachelors</td>
<td>Married-civ-spouse</td>
<td>Prof-specialty</td>
<td>Wife</td>
<td>White</td>
<td>False</td>
<td>41.000000</td>
<td>75409.001182</td>
<td>13.0</td>
<td>&gt;=50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">1</td>
<td>Private</td>
<td>Some-college</td>
<td>Never-married</td>
<td>Craft-repair</td>
<td>Not-in-family</td>
<td>White</td>
<td>False</td>
<td>24.000000</td>
<td>38455.005013</td>
<td>10.0</td>
<td>&lt;50k</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">2</td>
<td>Private</td>
<td>Assoc-acdm</td>
<td>Married-civ-spouse</td>
<td>Prof-specialty</td>
<td>Husband</td>
<td>White</td>
<td>False</td>
<td>48.000000</td>
<td>101299.003093</td>
<td>12.0</td>
<td>&lt;50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">3</td>
<td>Private</td>
<td>HS-grad</td>
<td>Never-married</td>
<td>Other-service</td>
<td>Other-relative</td>
<td>Black</td>
<td>False</td>
<td>42.000000</td>
<td>227465.999281</td>
<td>9.0</td>
<td>&lt;50k</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">4</td>
<td>State-gov</td>
<td>Some-college</td>
<td>Never-married</td>
<td>Prof-specialty</td>
<td>Not-in-family</td>
<td>White</td>
<td>False</td>
<td>20.999999</td>
<td>258489.997130</td>
<td>10.0</td>
<td>&lt;50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">5</td>
<td>Local-gov</td>
<td>12th</td>
<td>Married-civ-spouse</td>
<td>Tech-support</td>
<td>Husband</td>
<td>White</td>
<td>False</td>
<td>39.000000</td>
<td>207853.000067</td>
<td>8.0</td>
<td>&lt;50k</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">6</td>
<td>Private</td>
<td>Assoc-voc</td>
<td>Married-civ-spouse</td>
<td>Sales</td>
<td>Husband</td>
<td>White</td>
<td>False</td>
<td>36.000000</td>
<td>238414.998930</td>
<td>11.0</td>
<td>&gt;=50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">7</td>
<td>Private</td>
<td>HS-grad</td>
<td>Never-married</td>
<td>Craft-repair</td>
<td>Not-in-family</td>
<td>White</td>
<td>False</td>
<td>19.000000</td>
<td>445727.998937</td>
<td>9.0</td>
<td>&lt;50k</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">8</td>
<td>Local-gov</td>
<td>Bachelors</td>
<td>Married-civ-spouse</td>
<td>#na#</td>
<td>Husband</td>
<td>White</td>
<td>True</td>
<td>59.000000</td>
<td>196013.000174</td>
<td>10.0</td>
<td>&gt;=50k</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">9</td>
<td>Private</td>
<td>HS-grad</td>
<td>Married-civ-spouse</td>
<td>Prof-specialty</td>
<td>Wife</td>
<td>Black</td>
<td>False</td>
<td>39.000000</td>
<td>147500.000403</td>
<td>9.0</td>
<td>&lt;50k</td>
</tr>
</tbody>
</table>

我们可以使用 `tabular_learner`
方法定义一个模型。在我们定义模型时，`fastai` 将根据我们之前的 `y_names`
尝试推断损失函数。

**注意**：有时在处理表格数据时，您的 `y` 可能会被编码（例如 0 和
1）。在这种情况下，您应该在构造函数中显式传递
`y_block = CategoryBlock`，以便 `fastai` 不会假设您是在进行回归。

``` python
learn = tabular_learner(dls, metrics=accuracy)
```

我们可以使用 `fit_one_cycle` 方法训练该模型（`fine_tune`
方法在这里没有用处，因为我们没有预训练模型）。

``` python
learn.fit_one_cycle(1)
```

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: left;">
<th data-quarto-table-cell-role="th">epoch</th>
<th data-quarto-table-cell-role="th">train_loss</th>
<th data-quarto-table-cell-role="th">valid_loss</th>
<th data-quarto-table-cell-role="th">accuracy</th>
<th data-quarto-table-cell-role="th">time</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td>0</td>
<td>0.369360</td>
<td>0.348096</td>
<td>0.840756</td>
<td>00:05</td>
</tr>
</tbody>
</table>

我们可以看看一些预测：

``` python
learn.show_results()
```

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: right;">
<th data-quarto-table-cell-role="th"></th>
<th data-quarto-table-cell-role="th">workclass</th>
<th data-quarto-table-cell-role="th">education</th>
<th data-quarto-table-cell-role="th">marital-status</th>
<th data-quarto-table-cell-role="th">occupation</th>
<th data-quarto-table-cell-role="th">relationship</th>
<th data-quarto-table-cell-role="th">race</th>
<th data-quarto-table-cell-role="th">education-num_na</th>
<th data-quarto-table-cell-role="th">age</th>
<th data-quarto-table-cell-role="th">fnlwgt</th>
<th data-quarto-table-cell-role="th">education-num</th>
<th data-quarto-table-cell-role="th">salary</th>
<th data-quarto-table-cell-role="th">salary_pred</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td data-quarto-table-cell-role="th">0</td>
<td>5.0</td>
<td>12.0</td>
<td>3.0</td>
<td>8.0</td>
<td>1.0</td>
<td>5.0</td>
<td>1.0</td>
<td>0.324868</td>
<td>-1.138177</td>
<td>-0.424022</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">1</td>
<td>5.0</td>
<td>10.0</td>
<td>5.0</td>
<td>2.0</td>
<td>2.0</td>
<td>5.0</td>
<td>1.0</td>
<td>-0.482055</td>
<td>-1.351911</td>
<td>1.148438</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">2</td>
<td>5.0</td>
<td>12.0</td>
<td>6.0</td>
<td>12.0</td>
<td>3.0</td>
<td>5.0</td>
<td>1.0</td>
<td>-0.775482</td>
<td>0.138709</td>
<td>-0.424022</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">3</td>
<td>5.0</td>
<td>16.0</td>
<td>5.0</td>
<td>2.0</td>
<td>4.0</td>
<td>4.0</td>
<td>1.0</td>
<td>-1.362335</td>
<td>-0.227515</td>
<td>-0.030907</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">4</td>
<td>5.0</td>
<td>2.0</td>
<td>5.0</td>
<td>0.0</td>
<td>4.0</td>
<td>5.0</td>
<td>1.0</td>
<td>-1.509048</td>
<td>-0.191191</td>
<td>-1.210252</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">5</td>
<td>5.0</td>
<td>16.0</td>
<td>3.0</td>
<td>13.0</td>
<td>1.0</td>
<td>5.0</td>
<td>1.0</td>
<td>1.498575</td>
<td>-0.051096</td>
<td>-0.030907</td>
<td>1.0</td>
<td>1.0</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">6</td>
<td>5.0</td>
<td>12.0</td>
<td>3.0</td>
<td>15.0</td>
<td>1.0</td>
<td>5.0</td>
<td>1.0</td>
<td>-0.555412</td>
<td>0.039167</td>
<td>-0.424022</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">7</td>
<td>5.0</td>
<td>1.0</td>
<td>5.0</td>
<td>6.0</td>
<td>4.0</td>
<td>5.0</td>
<td>1.0</td>
<td>-1.582405</td>
<td>-1.396391</td>
<td>-1.603367</td>
<td>0.0</td>
<td>0.0</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">8</td>
<td>5.0</td>
<td>3.0</td>
<td>5.0</td>
<td>13.0</td>
<td>2.0</td>
<td>5.0</td>
<td>1.0</td>
<td>-1.362335</td>
<td>0.158354</td>
<td>-0.817137</td>
<td>0.0</td>
<td>0.0</td>
</tr>
</tbody>
</table>

或者在一行上使用预测方法：

``` python
row, clas, probs = learn.predict(df.iloc[0])
```

``` python
row.show()
```

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: right;">
<th data-quarto-table-cell-role="th"></th>
<th data-quarto-table-cell-role="th">workclass</th>
<th data-quarto-table-cell-role="th">education</th>
<th data-quarto-table-cell-role="th">marital-status</th>
<th data-quarto-table-cell-role="th">occupation</th>
<th data-quarto-table-cell-role="th">relationship</th>
<th data-quarto-table-cell-role="th">race</th>
<th data-quarto-table-cell-role="th">education-num_na</th>
<th data-quarto-table-cell-role="th">age</th>
<th data-quarto-table-cell-role="th">fnlwgt</th>
<th data-quarto-table-cell-role="th">education-num</th>
<th data-quarto-table-cell-role="th">salary</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td data-quarto-table-cell-role="th">0</td>
<td>Private</td>
<td>Assoc-acdm</td>
<td>Married-civ-spouse</td>
<td>#na#</td>
<td>Wife</td>
<td>White</td>
<td>False</td>
<td>49.0</td>
<td>101319.99788</td>
<td>12.0</td>
<td>&gt;=50k</td>
</tr>
</tbody>
</table>

``` python
clas, probs
```

    (tensor(1), tensor([0.4995, 0.5005]))

要对新的数据框进行预测，可以使用 `DataLoaders` 的 `test_dl`
方法。该数据框的列中不需要包含因变量。

``` python
test_df = df.copy()
test_df.drop(['salary'], axis=1, inplace=True)
dl = learn.dls.test_dl(test_df)
```

然后 `Learner.get_preds` 将会给你预测结果：

``` python
learn.get_preds(dl=dl)
```

    (tensor([[0.4995, 0.5005],
             [0.4882, 0.5118],
             [0.9824, 0.0176],
             ...,
             [0.5324, 0.4676],
             [0.7628, 0.2372],
             [0.5934, 0.4066]]), None)

<div>

> **Note**
>
> 由于机器学习模型无法神奇地理解从未训练过的类别，因此数据应反映这一点。如果您的测试数据中存在不同的缺失值，在训练之前应对此进行处理。

</div>

## `fastai`与其他库

如前所述，`TabularPandas`是一个强大且易于使用的表格数据预处理工具。与随机森林和XGBoost等库的集成只需一个额外的步骤，而`.dataloaders`调用为我们完成了这个步骤。让我们再次看看我们的`to`。它的值存储在一个类似于`DataFrame`的对象中，我们可以提取其中的`cats`、`conts`、`xs`和`ys`，如果我们需要的话：

``` python
to.xs[:3]
```

<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }
&#10;    .dataframe tbody tr th {
        vertical-align: top;
    }
&#10;    .dataframe thead th {
        text-align: right;
    }
</style>

<table class="dataframe" data-quarto-postprocess="true" data-border="1">
<thead>
<tr class="header" style="text-align: right;">
<th data-quarto-table-cell-role="th"></th>
<th data-quarto-table-cell-role="th">workclass</th>
<th data-quarto-table-cell-role="th">education</th>
<th data-quarto-table-cell-role="th">marital-status</th>
<th data-quarto-table-cell-role="th">occupation</th>
<th data-quarto-table-cell-role="th">relationship</th>
<th data-quarto-table-cell-role="th">race</th>
<th data-quarto-table-cell-role="th">education-num_na</th>
<th data-quarto-table-cell-role="th">age</th>
<th data-quarto-table-cell-role="th">fnlwgt</th>
<th data-quarto-table-cell-role="th">education-num</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td data-quarto-table-cell-role="th">25387</td>
<td>5</td>
<td>16</td>
<td>3</td>
<td>5</td>
<td>1</td>
<td>5</td>
<td>1</td>
<td>0.471582</td>
<td>-1.467756</td>
<td>-0.030907</td>
</tr>
<tr class="even">
<td data-quarto-table-cell-role="th">16872</td>
<td>1</td>
<td>16</td>
<td>5</td>
<td>1</td>
<td>4</td>
<td>5</td>
<td>1</td>
<td>-1.215622</td>
<td>-0.649792</td>
<td>-0.030907</td>
</tr>
<tr class="odd">
<td data-quarto-table-cell-role="th">25852</td>
<td>5</td>
<td>16</td>
<td>3</td>
<td>5</td>
<td>1</td>
<td>5</td>
<td>1</td>
<td>1.865358</td>
<td>-0.218915</td>
<td>-0.030907</td>
</tr>
</tbody>
</table>

</div>

现在一切都已经编码完成，您可以通过提取训练集和验证集及其值，将其发送给
XGBoost 或随机森林：

``` python
X_train, y_train = to.train.xs, to.train.ys.values.ravel()
X_test, y_test = to.valid.xs, to.valid.ys.values.ravel()
```

现在我们可以直接发送这个了！
